/*Lattice Gas Automaton
Initialisation:
Hier finden sich die Routinen zur Reservierung
der Systemressourcen sowie zur Erstellung und
"Bevoelkerung" der Matrix.
*/

#define DEBUG__nein

#include <stdio.h>
#include <stdlib.h>
#include "p5-lat.h"
#ifdef _UNIX_
#include <memory.h>
/*#define RAND_MAX 0xffffffffL    /* keine optimale Methode, geht aber */
#endif
/*#include <malloc.h>*/
#ifdef _MSDOS_
#include <mem.h>
#endif
#include <time.h>

extern int localSlaveNumber;
TZelle  *ptrAufFeldBasis;


TWord   wGrenzwert;
TWord   wFlussRechts;
TWord   wFlussLinks;

/* An dieser Stelle sind die Eingaben bereits gemacht worden,
d.h. die "Basisvariablen" sind bereits belegt/initialisiert. */

int     reserviereSpeicher(void);
int     bevoelkereZellen(void);
int     leereBarrieren(void);
int     leereRechteck(TWord,TWord,TWord,TWord);
int     Aufraeumen(void);

int     latticeInit(void)
{
int     res;
time_t  nichts;
float   zwwert;

        if(wSeed)               /* Zufallsgenerator initialisieren */
                srand(wSeed);
        else
        {
                res = time(&nichts);
                srand(res);
        }
        zwwert          = wDichte / 100.0;
        zwwert         *= RAND_MAX;
        wGrenzwert      = (TWord) zwwert;
        zwwert          = wFluss / 100.0;
        zwwert         *= RAND_MAX;
        wFlussRechts    = wGrenzwert - ((TWord) zwwert);
        wFlussLinks     = wGrenzwert + ((TWord) zwwert);
        res = reserviereSpeicher();
        if(res)
        {
                fprintf(stderr,"\nLATINIT: Fehler bei Speicherreservierung!(%d)",res);
                fflush(stderr);
                return(res);
        }
        res = bevoelkereZellen();
        if(res)
        {
                fprintf(stderr,"\nLATINIT: Fehler beim Bevoelkern!(%d) UNLOGISCH!!!!",res);
                fflush(stderr);
                return(res);
        }
/*TODO:
Barrieren beim Bevoelkern aussparen oder danach leeren
*/
        res = leereBarrieren();
        if(res)
        {
                fprintf(stderr,"\nLATINIT: Fehler beim Leeren der Barrieren ! UNLOGISCH!!!");
                fflush(stderr);
                return(res);
        }
        return(0);
}

int     reserviereSpeicher(void)
{
size_t  MemNeeded,hz,hs;
TDWord  MN2,hz2,hs2;
TZelle  *hPtr;

        hs        = wSpalten;           /* wegen 16-Bit Integer auf DOS-PCs */
        hz        = wZeilen + 2;
        MemNeeded = hz * hs;
        hz2     = hz;
        hs2     = hs;
        MN2     = hz2 * hs2 - 2;
        if(MN2 > MemNeeded)
        {
                fprintf(stderr,"\nLATINIT: FEHLER !! Ueberlauf (MemNeeded)");
                fflush(stderr);
                exit(1);
        }
        fprintf(stderr,"\nLATINIT: Allokiere %ld Bytes ...",MN2+1);
        fflush(stderr);
        ptrAufFeldBasis = malloc(MemNeeded);
        if(ptrAufFeldBasis==NULL)
	  {
	      ptrZellFeldBasis = NULL;
                return(1);
	  }
        else
	  {
	      hPtr = ptrAufFeldBasis;
	      hPtr += wSpalten;
	      ptrZellFeldBasis = hPtr;
	      return(0);
	  }
}

int     bevoelkereZellen(void)
{
TWord           runx,runy;
TRichtung       rund;
TZelle          *ptrHZelle;     /* Hilfzeiger auf bearbeitete Zelle */

        ptrHZelle       = (TZelle *) ptrZellFeldBasis;
        fprintf(stderr,"\nBeginne mit Initialisierung der Matrix\n");
        for(runx=0;runx<wSpalten;runx++)
        {
                fprintf(stderr,".");
                fflush(stderr);
                for(runy=0;runy<wZeilen;runy++)
                {
                        *ptrHZelle = 0; /* Zelle loeschen */
                        for(rund=ObenRechts;rund<UntenLinks;rund++)
                        {       /* Fluss nach Rechts setzen */
                                if(rand() < wFlussRechts)
                                        *ptrHZelle      |= (0x01 << rund);
                        }
                        for(rund=UntenLinks;rund<STOP;rund++)
                        {       /* Fluss n. Links setzen */
                                if(rand() < wFlussLinks)
                                        *ptrHZelle      |= (0x01 << rund);
                        }
                        ptrHZelle++;
                }
        }
        return(0);
}

int     leereBarrieren(void)
{
TBListe *ptrH;
int res;

        ptrH    = BListeWurzel;
        while(ptrH->next != NULL)
        {
                ptrH    = ptrH->next;
                res = leereRechteck(ptrH->Knoten->LinksOben.x,ptrH->Knoten->LinksOben.y,ptrH->Knoten->RechtsUnten.x,ptrH->Knoten->RechtsUnten.y);
                if(res)
                {
                        fprintf(stderr,"\nLATINIT: Fehler!! (leereBarrieren)");
                        fflush(stderr);
                        return(res);
                }
        }
        return(0);
}

int     leereRechteck(TWord x1,TWord y1,TWord x2,TWord y2)
{
TZelle  *ptrH,*ptrH2;
TWord   runx,runy;

#ifdef DEBUG__
fprintf(stderr,"PN(%d), (%d,%d)x(%d,%d)\n",localSlaveNumber,x1,y1,x2,y2);
fflush(stderr);
sleep(5);
#endif
        ptrH    = (TZelle *) ptrZellFeldBasis;
        ptrH   += x1;
        ptrH   += (y1 * wSpalten);      /* Adresse der linken oberen Ecke */
        ptrH2   = ptrH;
        for(runy=y1;runy<=y2;runy++)
        {
                for(runx=x1;runx<=x2;runx++)
                {
/*TODO: eigentlich muesste man ja noch die Richtungen beruecksichtigen,
aber das kommt spaeter*/
                        *ptrH2 = 0;
                        ptrH2++;
                }
                ptrH   += wSpalten;
                ptrH2   = ptrH;         /* naechste Zeile */
        }
        return(0);
}

int     Aufraeumen(void)
{
        free(ptrAufFeldBasis);
        return(0);
}








