
/* Masterprogramm zum Lattice Gas Automaten */
/* Transputerpraktikum SS95 Praktikumsgruppe 5 */

#define WARTEN__nein
#define CALC__          /* erlaubt die Berechnung */
#define AUSGABE__
#define IODEBUG__nein
#define DEBUG__nein
#define _MASTER_
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "PsxIO.h"
#include "slavecom.h"
/*#include "p5-lat.h"*/

extern
int readFile(char *);
extern
int latticeShow(void);

int ReadLatticeInformationFile(char *);
int VerteileAufgaben(void);
int Berechne(void);
int reserviereSpeicher(void);
int LeseBilddaten(void);

void ResetSlaves(void);
int  ZaehleSlaves(void);
int  SendToSlave(int,char *,unsigned);
int  ReceiveFromSlave(int,char *,unsigned);
void SendBarrierenToSlave(int);

TKommando Kommando;
int slaveAnzahl;

int main(int argc,char *argv[])
{
int fehler;


fprintf(stderr,"ARGC(%d) ARGV[1](%s)\n",argc,argv[1]);
fflush(stderr);
    if(argc < 2)
      fehler = ReadLatticeInformationFile("lattice.dat");
    else
      fehler = ReadLatticeInformationFile(argv[1]);
    if(fehler)
        {
            fprintf(stderr,"Fehler beim Lesen der Beschreibungsdatei\n");
            fflush(stderr);
            ResetSlaves();
            exit(1);
        }
    printf("Beschreibungsdatei gelesen.\n");
    fflush(stdout);
#ifdef WARTEN__
sleep(2);
#endif
    fehler = VerteileAufgaben();
#ifdef WARTEN__
sleep(2);
#endif
    if(fehler)
        {
            fprintf(stderr,"Fehler beim Verteilen der Aufgaben\n");
            fflush(stderr);
            ResetSlaves();
            exit(1);
        }
    fflush(stdout);
    printf("Aufgaben verteilt.\n");
    fflush(stdout); 
#ifdef CALC__
    fehler = Berechne();
    if(fehler)
        {
            fprintf(stderr,"Fehler beim Berechnen\n");
            fflush(stderr);
            ResetSlaves();
            exit(1);
        }
    fflush(stdout);
#endif
    printf("Na also, klappte ja wunderbar ;-))))\n");
    ResetSlaves();
    printf("Slaves beendet !\n");
    fflush(stdout);
    return(0);
}

 
int ReadLatticeInformationFile(char *dateiname)
{
/* Kopie d. alten benutzen ! (lesen + initialisieren, Matrix dezentral halten)*/
  int res = readFile(dateiname);  /* def. in LATMAIN */
  return(res);
}

int VerteileAufgaben(void)
{
  int runvar;

  slaveAnzahl = ZaehleSlaves();
  printf("%2d Slaves gefunden. Versende Aufgaben...\n",slaveAnzahl);
  Kommando.Data.beschreibung.slaves = slaveAnzahl;
/* Initialisierung d. Kommandostruktur */
  Kommando.code = BESCHREIBUNG;
  Kommando.Data.beschreibung.zeilen = wZeilen;
  Kommando.Data.beschreibung.spalten = wSpalten;
  Kommando.Data.beschreibung.barrieren = wBarrieren;
  Kommando.Data.beschreibung.seed = wSeed;
  Kommando.Data.beschreibung.dispX = wBildZeile;
  Kommando.Data.beschreibung.dispY = wBildSpalte;
  Kommando.Data.beschreibung.dispSteps = wBildSchritte;
  Kommando.Data.beschreibung.steps = wMaxZeit;
  Kommando.Data.beschreibung.dichte = wDichte;
  Kommando.Data.beschreibung.fluss = wFluss;
  SendToSlave(RING_LO,(char *) &Kommando,sizeof(TKommando));
  ReceiveFromSlave(RING_RI,(char *) &Kommando,sizeof(TKommando));
  if(Kommando.code == FEHLER)
  {
      fprintf(stderr,"FEHLER !!!!! beim Verteilen!");
      fflush(stderr);
      ResetSlaves();
      exit(1);
  }
/* ACHTUNG!: sicherstellen, dass Aufgabenfeld richtig gesetzt ist */
/* OK, Aufgaben sind verteilt, nun noch die Barrieren uebertragen */
 /* for(runvar=0;runvar<slaveAnzahl;runvar++)
  */
  SendBarrierenToSlave(runvar);
  return(0);
}

int Berechne(void)
{
  int res = reserviereSpeicher();
  int runvar;
  clock_t zeita,zeitn,zeits,zeits2;

  if(res)
  {
      fprintf(stderr,"FEHLER, zuwenig Speicher !!\n");
      fflush(stderr);
      ResetSlaves();
      exit(1);
  }      
  Kommando.code = RECHNE;
  Kommando.Data.schritte = wMaxZeit;
  SendToSlave(RING_LO,(char *) &Kommando,sizeof(TKommando));
  ReceiveFromSlave(RING_RI,(char *) &Kommando,sizeof(TKommando));
/* OK, Kommando zur Berechnung abgesetzt */
  latticeShow();  /* zwecks Initialisierung d. Grafikausgabe */
  zeitn = clock();
  zeits = zeitn;
  zeits2 = 0;   /* ohne Initialisierung */
  for(runvar=0;runvar<wMaxZeit;runvar+=wBildSchritte)
  {
      zeita = zeitn;
#ifdef AUSGABE__
      res = LeseBilddaten();
#endif
      if(res)
      {
          fprintf(stderr,"Master: Fehler beim Lesen der Bilddaten\n");
          fflush(stderr);
          /*ResetSlaves();*/ /* wg. Slave am Direktkanal */
          exit(1);
      }
#ifdef AUSGABE__
      res = latticeShow();
#endif
      if(res)
      {
          fprintf(stderr,"Fehler bei der X11-Ausgabe\n");
          fflush(stderr);
          /*ResetSlaves();*/ /* s.o. */
          exit(1);
      }
      zeitn = clock();
      if(zeits2==0) zeits2 = zeitn;
      printf("Master: Zeit: (%lf)\n",(double) (zeitn-zeita)/CLK_TCK);
      fflush(stdout);
  }
  printf("Master, Gesamtzeit: (%lf),(%lf)\n",(double) (zeitn-zeits)/CLK_TCK,(double) (zeitn-zeits2)/CLK_TCK);
  fflush(stdout);
  return(0);
}

/* Slave - Kommunikation, evtl. auslagern */

void ResetSlaves(void)
{
/* evtl. in Slave Subdatei auslagern */
int rv;

  Kommando.code = RESET;
  SendToSlave(RING_LO,(char *) &Kommando,sizeof(TKommando));
  ReceiveFromSlave(RING_RI,(char *) &Kommando,sizeof(TKommando));
}

int ZaehleSlaves(void)
{
int result;

  Kommando.code = COUNT;
  Kommando.Data.zaehler = 0;
  result = SendToSlave(RING_LO,(char *) &Kommando,sizeof(TKommando));
  if(result == -1)
    {
        fprintf(stderr,"Fehler (STS - Count)\n");
        fflush(stderr);
        exit(1);
    }
  result = ReceiveFromSlave(RING_RI,(char *) &Kommando,sizeof(TKommando));
  if(result == -1)
    {
        fprintf(stderr,"Fehler (RFS - Count)\n");
        fflush(stderr);
        exit(1);
    }
  result = Kommando.Data.zaehler;
  return(result);
}

int SendToSlave(int channel,char *address,unsigned length)
{
#ifdef DEBUG__
#ifdef IODEBUG__
fprintf(stderr,"MASTER: schreibe auf Kanal(%d)\n",channel);
fflush(stderr);
#endif
#endif
    return(PsxWrite(channel,address,length));
}

int ReceiveFromSlave(int channel,char *address,unsigned length)
{
#ifdef DEBUG__
#ifdef IODEBUG__
fprintf(stderr,"MASTER: lese von Kanal (%d)\n",channel);
fflush(stderr);
#endif
#endif
    return(PsxRead(channel,address,length));
}

void SendBarrierenToSlave(int slaveno)
{
    TBListe *hptr;

    hptr = BListeWurzel->next;
    while(hptr != NULL)
    {
        Kommando.code = BARRIERE;
        Kommando.Data.barriere.loX = hptr->Knoten->LinksOben.x;
        Kommando.Data.barriere.loY = hptr->Knoten->LinksOben.y;
        Kommando.Data.barriere.ruX = hptr->Knoten->RechtsUnten.x;
        Kommando.Data.barriere.ruY = hptr->Knoten->RechtsUnten.y;
        SendToSlave(RING_LO,(char *) &Kommando,sizeof(TKommando));
#ifdef DEBUG__
fprintf(stderr,"Master(Barriere): (%d,%d)x(%d,%d)\n",Kommando.Data.barriere.loX,Kommando.Data.barriere.loY,Kommando.Data.barriere.ruX,Kommando.Data.barriere.ruY);
fflush(stderr);
#endif
        ReceiveFromSlave(RING_RI,(char *) &Kommando,sizeof(TKommando));
        if(Kommando.code == FEHLER)
        {
            fprintf(stderr,"FEHLER beim Senden der Barrieren !\n");
            fflush(stderr);
            ResetSlaves();
        }
        hptr = hptr->next;
    }
}

int     reserviereSpeicher(void)
{
size_t  MemNeeded,hz,hs;
TDWord  MN2,hz2,hs2;

        hs        = wSpalten;           /* wegen 16-Bit Integer auf DOS-PCs */
        hz        = wZeilen+slaveAnzahl; /* Sicherheitspuffer */
        MemNeeded = hz * hs;
        hz2     = hz;
        hs2     = hs;
        MN2     = hz2 * hs2 - 1;
        if(MN2 > MemNeeded)
        {
                fprintf(stderr,"\nFEHLER !! Ueberlauf (MemNeeded)");
                fflush(stderr);
                exit(1);
        }
        fprintf(stderr,"\nAllokiere %ld Bytes ...",MN2+1);
        fflush(stderr);
        ptrZellFeldBasis = malloc(MemNeeded);
        if(ptrZellFeldBasis==NULL)
                return(1);
        else    return(0);
}

int LeseBilddaten(void)
{
    int runvar;
    TZelle *zptr;
    int length;
#ifdef AUSGABE__
    zptr = (TZelle *) ptrZellFeldBasis;
    length = wSpalten*((wZeilen+slaveAnzahl-1)/slaveAnzahl);
fprintf(stderr,"MASTER: Length (%d)\n",length);
fflush(stderr);   
    for(runvar=0;runvar<slaveAnzahl;runvar++)
    {
        ReceiveFromSlave(DIREKT_I(runvar),(char *) zptr,length);
        zptr = zptr + length;
    }
    printf("Master: Bilddaten gelesen.\n");
    fflush(stdout);
#endif
    return(0);
}
       

int Dichte(TZelle *zp)
{/*TODO rausschmeisen und vernuenftig deklarieren */
	return(6);
}







